{
    Файл является частью библиотеки RubiRobotLib

    © 2020-2021 Слинкин Д.А.

    Модуль алгоритмов библиотеки RubiRobotLib v0.2.6+

    Подробная информация о лицензировании находится в файле README,
    входящем в состав библиотеки.

    ДАННОЕ ПРОГРАММНОЕ ОБЕСПЕЧЕНИЕ ПРЕДОСТАВЛЯЕТСЯ «КАК ЕСТЬ», БЕЗ
    КАКИХ-ЛИБО ГАРАНТИЙ, ЯВНО ВЫРАЖЕННЫХ ИЛИ ПОДРАЗУМЕВАЕМЫХ,
    ВКЛЮЧАЯ ГАРАНТИИ ТОВАРНОЙ ПРИГОДНОСТИ, СООТВЕТСТВИЯ ПО ЕГО
    КОНКРЕТНОМУ НАЗНАЧЕНИЮ И ОТСУТСТВИЯ НАРУШЕНИЙ, НО НЕ ОГРАНИЧИВАЯСЬ 
    ИМИ. НИ В КАКОМ СЛУЧАЕ АВТОРЫ ИЛИ ПРАВООБЛАДАТЕЛИ НЕ НЕСУТ
    ОТВЕТСТВЕННОСТИ ПО КАКИМ-ЛИБО ИСКАМ, ЗА УЩЕРБ ИЛИ ПО ИНЫМ ТРЕБОВАНИЯМ,
    В ТОМ ЧИСЛЕ, ПРИ ДЕЙСТВИИ КОНТРАКТА, ДЕЛИКТЕ ИЛИ ИНОЙ СИТУАЦИИ,
    ВОЗНИКШИМ ИЗ-ЗА ИСПОЛЬЗОВАНИЯ ПРОГРАММНОГО ОБЕСПЕЧЕНИЯ ИЛИ ИНЫХ
    ДЕЙСТВИЙ С ПРОГРАММНЫМ ОБЕСПЕЧЕНИЕМ.

 **********************************************************************}


{$I uev3defines.inc}
unit rubiroAlgo;

{$modeswitch nestedprocvars}
interface
uses rubiroMotors,rubiroSensors;


{

  Движение по гироскопу с использованием одного мотора.
  Может использоваться, например, в роботах на базе "пятиминутки" для
  поворота вокруг одного колеса.

  Запуск мотора motor (если мотор не указан, используется первый большой мотор в системе)
  с максимальной скоростью maxSpeed и минимальной скоростью
  minSpeed с остановкой по достижению угла targetAngle гироскопа
  gyro (если гироскоп не указан, используется первый гироскоп в системе).
  minSpeed и maxSpeed должны быть не равны 0 и всегда одного знака, который
  определяет направление движения мотора.
  Параметр correct обеспечивает автоопределение корректности движения, а
  именно: если при движении происходит удаление от целевого угла поворота
  targetAngle более чем на correct градусов относительно первого замера,
  то мотор запускается в противоположную сторону и correct обнуляется.
  Автоопределение отключено при correct=0.
  Параметр koef (коэффициент пропорционального регулятора)
  определяет плавность замедления движения мотора при
  приближении к целевому углу.

  Примеры запуска:
    Запускается второй большой мотор на максимальной скорости 30, минимальной 10,
    до достижения угла поворота первого гироскопа в значение 180.
    Параметр автоопределения корректности движения не указан, следовательно
    автоопределение включено и равно 5 градусам, то есть если в результате запуска
    мотора на заданной скорости текущее значение угла поворота гироскопа вместо
    приближения удалится от целевого угла более чем на 5 градусов, движение мотора
    будет инвертировано.
    Параметр koef не указан, значит равен 1
    Парметр stopmotors не указан, значит моторы будут остановлены по окончании
    работы процедуры

    TurnGyroWheel(ev3gyro, ev3motor2, 180, 30, 10);

    Запускается первый большой мотор на максимальной скорости 30, минимальной 20,
    до достижения угла поворота первого гироскопа в значение -90.
    Остальные параметры вызова процедуры аналогичны предыдущему примеру.

    TurnGyroWheel(-90, 50, 20);
}
procedure TurnGyroWheel(
                        gyro:TEv3GyroSensor;
                        motor:TEv3TachoMotor;
                        targetAngle:integer;
                        speedMax,speedMin:double;
                        correct:integer=5; koef:double=1;
                        stopmotor:boolean=true
                        );

procedure TurnGyroWheel(
                        targetAngle:integer;
                        speedMax,speedMin:double;
                        correct:integer=5; koef:double=1;
                        stopmotor:boolean=true
                        );

{
Движение по гироскопу gyro с использованием двух моторов LMotor и RMotor.

Может использоваться, например, в роботах на базе "пятиминутки" для
поворота вокруг оси.

Если гироскоп не указан, используется первый гироскоп в системе.
Если моторы не указаны, то предполагается штатное подключение двух больших моторов:
левого - к порту с меньшим номером,
правого - к порту с большим номером

targetAngle - целевой угол, на который будет остуановлен гироскоп по окончании поворота
maxspeed - максимальная скорость поворота, всегда >0
minspeed - минимальная скорость поворота, всегда >0
koef - коэффициент пропорционального регулятора,
       определяет плавность замедления движения мотора при
       приближении к целевому углу.
stopmotors - флаг остановки моторов по окончании работы процедуры

Движение всегда направлено к целевому углу.
}
procedure TurnGyroAxis(
                      gyro:TEv3GyroSensor;
                      LMotor,RMotor:TEv3TachoMotor;
                      targetAngle:integer;
                      maxspeed,minspeed:double;
                      koef:double=1;
                      stopmotors:boolean=true
                      );

procedure TurnGyroAxis(
                      targetAngle:integer;
                      maxspeed,minspeed:double;
                      koef:double=1;
                      stopmotors:boolean=true
                      );


{
 Прямолинейное движение по гироскопу gyro на расстояние distance
 с использованием двух моторов LMotor и RMotor.

 Если гироскоп не указан, используется первый гироскоп в системе.
 Если моторы не указаны, то предполагается штатное подключение двух больших моторов:
 левого - к порту с меньшим номером,
 правого - к порту с большим номером

 targetAngle - целевой угол, к установке которого будет стремиться робот во время
 движения
 speed - скорость движения, положительное значение для движения вперед,
         отрицательное - назад
 distance - расстояниев градусах, которое должен преодолеть робот до остановки
 koef - коэффициент пропорционального регулятора
 stopmotors - флаг остановки моторов по окончании работы процедуры
}
procedure RunGyroDirectDistance(
                          gyro:TEv3GyroSensor;
                          LMotor,RMotor:TEv3TachoMotor;
                          targetAngle:integer; speed:double;
                          distance:double;
                          koef:double=1;
                          stopmotors:boolean=true
                          );
procedure RunGyroDirectDistance(
                          targetAngle:integer; speed:double;
                          distance:double;
                          koef:double=1;
                          stopmotors:boolean=true
                          );
{
 Прямолинейное движение по гироскопу на заданное расстояние
 с использованием двух моторов.

 Процедура специально подготовлена для параллельного запуска, но может быть
 вызвана и непосредственно.

 В качестве единственного параметра передается вариантный массив
 Параметры в массив передаются в том-же порядке, как и в RunGyroDirectDistance.
 Если первые параметры - объекты, они считаются, соответственно, гироскопом,
 левым и правым мотором.
}
procedure RunGyroDirectDistanceForParallel(data:array of const);



{
 Прямолинейное движение по гироскопу gyro до черной линии с максимальным
 уровнем отраженного света maxReflect,
 с использованием двух моторов LMotor и RMotor.

 Если гироскоп не указан, используется первый гироскоп в системе.
 Если моторы не указаны, то предполагается штатное подключение двух больших моторов:
 левого - к порту с меньшим номером,
 правого - к порту с большим номером
 Если не указаны оба датчика цвета, то будут использоваться первых два датчика
 цвета. Если не указан один из датчиков, то он будет проигнорирован.

 targetAngle - целевой угол, к установке которого будет стремиться робот во время
 движения
 speed - скорость движения, положительное значение для движения вперед,
         отрицательное - назад
 maxReflect - максимальное значение уровня отраженного цвета. Если суммарное
              значение отраженного света, полученное от датчиков цвета,
              меньше данной величины, процедура завершает
              свою работу.
 koef - коэффициент пропорционального регулятора
 stopmotors - флаг остановки моторов по окончании работы процедуры
}

procedure RunGyroDirectToBlackLine(
                          gyro:TEv3GyroSensor;
                          LMotor,RMotor:TEv3TachoMotor;
                          LColor,RColor:TEv3ColorSensor;
                          targetAngle:integer; speed:double;
                          MaxReflect:double;
                          koef:double=1;
                          stopmotors:boolean=true
                          );

procedure RunGyroDirectToBlackLine(
                          Color:TEv3ColorSensor;
                          targetAngle:integer; speed:double;
                          MaxReflect:double;
                          koef:double=1;
                          stopmotors:boolean=true
                          );

procedure RunGyroDirectToBlackLine(
                          targetAngle:integer; speed:double;
                          MaxReflect:double;
                          koef:double=1;
                          stopmotors:boolean=true
                          );

// типы стоп-функций
type TStopLocalFunc=function ():boolean is nested;
     TStopGlobalFunc=function ():boolean;
     TStopObjectFunc=function ():boolean;

{
 Универсальная процедура прямолинейного движения по гироскопу gyro
 с использованием двух моторов LMotor и RMotor.

 Если гироскоп не указан, используется первый гироскоп в системе.
 Если моторы не указаны, то предполагается штатное подключение двух больших моторов:
 левого - к порту с меньшим номером,
 правого - к порту с большим номером

 stopfunc - ссылка на пользовательскую стоп-функцию, которая вызывается на каждой
 итерации цикла внутри RunGyro и должна возвратить истину, если RunGyro должна
 завершить работу.
 targetAngle - целевой угол, к установке которого будет стремиться робот во время
 движения
 speed - скорость движения, положительное значение для движения вперед,
         отрицательное - назад
 koef - коэффициент пропорционального регулятора
 stopmotors - флаг остановки моторов по окончании работы процедуры
}


procedure doRunGyro(
                    stopfunc:TStopLocalFunc;
                    gyro: TEv3GyroSensor;
                    LMotor, RMotor: TEv3TachoMotor;
                    targetAngle: integer;
                    speed: double;
                    koef:double;
                    stopmotors:boolean=true
                    );

procedure doRunGyro(
                    stopfunc:TStopGlobalFunc;
                    gyro: TEv3GyroSensor;
                    LMotor, RMotor: TEv3TachoMotor;
                    targetAngle: integer;
                    speed: double;
                    koef:double;
                    stopmotors:boolean=true
                    );

procedure doRunGyro(
                    stopfunc:TStopObjectFunc;
                    gyro: TEv3GyroSensor;
                    LMotor, RMotor: TEv3TachoMotor;
                    targetAngle: integer;
                    speed: double;
                    koef:double;
                    stopmotors:boolean=true
                    );



implementation
 uses uev3,rubiroParallel;

 __WARNGLOBALOFF

procedure TurnGyroWheel(gyro: TEv3GyroSensor; motor: TEv3TachoMotor;
  targetAngle: integer; speedMax, speedMin: double; correct: integer;
  koef: double; stopmotor: boolean);
  var
      err,olderr:double;
      signAngle,signSpeed:integer;
      deltaspeed:double;
 begin
  try
   if not assigned(motor) then motor:=ev3Motor;
   if not assigned(gyro) then gyro:=ev3gyro;
   if speedMin*speedMax<0 then exit;
   if speedMax>0 then signSpeed:=1 else signSpeed:=-1;
   speedMax:=abs(speedMax); speedMin:=abs(speedMin);
   err:=gyro.angle-targetAngle;
   olderr:=err;
   if err<0 then signAngle:=-1 else if err>0 then signAngle:=1 else exit;
   while true do begin
    err:=gyro.angle-targetAngle;
    if err*signAngle<=0 then exit;
    if (correct>0) and ((abs(err)-abs(olderr))>correct) then begin
      signSpeed:=-signSpeed;
      correct:=0;
    end;
    deltaspeed:=abs(err*koef);
    if deltaspeed>speedMax then deltaspeed:=signSpeed*speedMax
    else
    if deltaspeed<speedMin then deltaspeed:=signSpeed*speedMin
    else deltaspeed:=signSpeed*deltaspeed;
    motor.run(deltaspeed);
   end;
  finally
   if stopmotor then motor.stop();
  end;
 end;

procedure TurnGyroWheel(targetAngle: integer; speedMax, speedMin: double;
  correct: integer; koef: double; stopmotor: boolean);
begin
 TurnGyroWheel(nil,nil,targetAngle,speedMax, speedMin, correct, koef, stopmotor);
end;


procedure TurnGyroAxis(gyro: TEv3GyroSensor; LMotor, RMotor: TEv3TachoMotor;
  targetAngle: integer; maxspeed, minspeed: double; koef: double;
  stopmotors: boolean);
 var
     err:double;
     sign:integer;
     deltaspeed:double;
begin
 maxspeed:=abs(maxspeed);
 minspeed:=abs(minspeed);
 try
  if gyro=nil then gyro:=ev3gyro;
  if lmotor=nil then lmotor:=ev3Motor1;
  if rmotor=nil then rmotor:=ev3Motor2;
  sign:=gyro.angle-targetAngle;
  while true do begin
   err:=gyro.angle-targetAngle;
   if err*sign<=0 then exit;
   deltaspeed:=err*koef;
   if abs(deltaspeed)>maxspeed then
    if deltaspeed<0 then deltaspeed:=-maxspeed else deltaspeed:=maxspeed;
   if deltaspeed<0 then deltaspeed:=deltaspeed-minspeed
   else deltaspeed:=deltaspeed+minspeed;
   lmotor.run(-deltaspeed);
   rmotor.run(deltaspeed);
  end;
 finally
  if stopmotors then begin
   lmotor.stop();
   rmotor.stop();
  end;
 end;
end;

procedure TurnGyroAxis(targetAngle: integer; maxspeed, minspeed: double;
  koef: double; stopmotors: boolean);
begin
 TurnGyroAxis(nil,nil,nil,targetAngle,maxspeed, minspeed, koef, stopmotors);
end;

procedure doRunGyro(
                    stopfunc:TStopLocalFunc;
                    gyro: TEv3GyroSensor;
                    LMotor, RMotor: TEv3TachoMotor;
                    targetAngle: integer;
                    speed: double;
                    koef:double;
                    stopmotors:boolean=true
                    );
 var
     err:integer;errold:integer=0;
 begin
  if stopfunc=nil then exit;
  try
   if gyro=nil then gyro:=ev3gyro;
   if lmotor=nil then lmotor:=ev3Motor1;
   if rmotor=nil then rmotor:=ev3Motor2;
   lmotor.run(speed);
   rmotor.run(speed);
   while true do begin
    //if (santim>0)and(motorast(9)>santim*cm) then break;
    //if pstop^=true then break;
    if stopfunc() then break;
    err:=gyro.angle-targetAngle;
    if err=errold then begin ev3.sleep(10); continue; end;
    lmotor.run(speed-koef*err);
    rmotor.run(speed+koef*err);
    errold:=err;
   end;
  finally
   if stopmotors then begin
    lmotor.stop();
    rmotor.stop();
   end;
  end;
 end;

procedure doRunGyro(stopfunc: TStopGlobalFunc; gyro: TEv3GyroSensor; LMotor,
  RMotor: TEv3TachoMotor; targetAngle: integer; speed: double; koef: double;
  stopmotors: boolean);
  function local():boolean; begin result:=stopfunc(); end;
begin
  if stopfunc=nil then exit;
  doRunGyro( @local, gyro, LMotor, RMotor,
             targetAngle, speed, koef, stopmotors);
end;

procedure doRunGyro(stopfunc: TStopObjectFunc; gyro: TEv3GyroSensor; LMotor,
  RMotor: TEv3TachoMotor; targetAngle: integer; speed: double; koef: double;
  stopmotors: boolean);
  function local():boolean; begin result:=stopfunc(); end;
begin
 if stopfunc=nil then exit;
 doRunGyro( @local, gyro, LMotor, RMotor,
            targetAngle, speed, koef, stopmotors);
end;



function motorast(id:integer; lmotor,rmotor:TEV3TachoMotor):double;
 begin
  result:=abs((lmotor.counter(cdeg,id)+rmotor.counter(cdeg,id))/2);
 end;

procedure motostart(id:integer; lmotor,rmotor:TEV3TachoMotor);
 begin
  lmotor.counterStart(id);
  rmotor.counterStart(id);
 end;

procedure RunGyroDirectDistance(gyro: TEv3GyroSensor; LMotor, RMotor: TEv3TachoMotor;
  targetAngle: integer; speed: double; distance: double; koef: double; stopmotors: boolean);
 function local():boolean; begin result:=motorast(9,lmotor,rmotor)>distance; end;
begin
 if lmotor=nil then lmotor:=ev3motor1;
 if rmotor=nil then rmotor:=ev3motor2;
 motostart(9,lmotor,rmotor);
 doRunGyro(@local, gyro, LMotor, RMotor,
  targetAngle, speed, koef, stopmotors);
end;

procedure RunGyroDirectDistance(targetAngle: integer; speed: double;
  distance: double; koef: double; stopmotors: boolean);
begin
 RunGyroDirectDistance(nil,nil,nil,targetAngle,speed,distance, koef, stopmotors);
end;

procedure RunGyroDirectDistanceForParallel(data: array of const);
 var gyro: TEv3GyroSensor=nil;
     LMotor: TEv3TachoMotor=nil;
     RMotor: TEv3TachoMotor=nil;
     targetAngle: integer=0; speed: double=20.0; distance: double=360.0;
     koef: double=1.0; stopmotors: boolean=true;
     current:integer=0;
begin
  if isObject(data[0]) then begin
   if length(data)>current then begin gyro:=toObject(data[current]) as TEv3GyroSensor;inc(current); end;
   if length(data)>current then begin lmotor:=toObject(data[current])as TEv3TachoMotor;inc(current); end;
   if length(data)>current then begin rmotor:=toObject(data[current])as TEv3TachoMotor;inc(current); end;
  end;
  if length(data)>current then begin targetAngle:=toInt(data[current]);inc(current); end;
  if length(data)>current then begin speed:=toReal(data[current]);inc(current); end;
  if length(data)>current then begin distance:=toReal(data[current]);inc(current); end;
  if length(data)>current then begin koef:=toReal(data[current]);inc(current); end;
  if length(data)>current then begin stopmotors:=toBool(data[current]);inc(current); end;
  RunGyroDirectDistance(gyro,lmotor,rmotor,targetAngle,speed,distance, koef, stopmotors);
end;


procedure RunGyroDirectToBlackLine(gyro: TEv3GyroSensor; LMotor,
  RMotor: TEv3TachoMotor; LColor, RColor: TEv3ColorSensor;
  targetAngle: integer; speed: double; MaxReflect: double; koef: double;
  stopmotors: boolean);
function local():boolean; begin result:=(lcolor.reflect+rcolor.reflect)<maxReflect; end;
begin
 if (LColor=nil)and(RColor=nil) then begin
  lcolor:=ev3color1; rcolor:=ev3color2;
 end
 else if (LColor=nil)and(RColor<>nil) then begin
  lcolor:=rcolor; maxReflect*=2;
 end
 else if (LColor<>nil)and(RColor=nil) then begin
  rcolor:=lcolor; maxReflect*=2;
 end;
 doRunGyro(@local, gyro, LMotor, RMotor,
  targetAngle, speed, koef, stopmotors);
end;

procedure RunGyroDirectToBlackLine(Color: TEv3ColorSensor;
  targetAngle: integer; speed: double; MaxReflect: double; koef: double;
  stopmotors: boolean);
begin
 RunGyroDirectToBlackLine(nil,nil,nil,Color,nil,targetAngle,speed,maxReflect, koef, stopmotors);
end;

procedure RunGyroDirectToBlackLine(targetAngle: integer; speed: double;
  MaxReflect: double; koef: double; stopmotors: boolean);
begin
 RunGyroDirectToBlackLine(nil,nil,nil,nil,nil,targetAngle,speed,maxReflect, koef, stopmotors);
end;


end.

