{
    Файл является частью библиотеки RubiRobotLib

    © 2017-2021 Слинкин Д.А.

    Обработка датчиков контроллера Lego EV3

    Подробная информация о лицензировании находится в файле README,
    входящем в состав библиотеки.

    ДАННОЕ ПРОГРАММНОЕ ОБЕСПЕЧЕНИЕ ПРЕДОСТАВЛЯЕТСЯ «КАК ЕСТЬ», БЕЗ
    КАКИХ-ЛИБО ГАРАНТИЙ, ЯВНО ВЫРАЖЕННЫХ ИЛИ ПОДРАЗУМЕВАЕМЫХ,
    ВКЛЮЧАЯ ГАРАНТИИ ТОВАРНОЙ ПРИГОДНОСТИ, СООТВЕТСТВИЯ ПО ЕГО
    КОНКРЕТНОМУ НАЗНАЧЕНИЮ И ОТСУТСТВИЯ НАРУШЕНИЙ, НО НЕ ОГРАНИЧИВАЯСЬ 
    ИМИ. НИ В КАКОМ СЛУЧАЕ АВТОРЫ ИЛИ ПРАВООБЛАДАТЕЛИ НЕ НЕСУТ
    ОТВЕТСТВЕННОСТИ ПО КАКИМ-ЛИБО ИСКАМ, ЗА УЩЕРБ ИЛИ ПО ИНЫМ ТРЕБОВАНИЯМ,
    В ТОМ ЧИСЛЕ, ПРИ ДЕЙСТВИИ КОНТРАКТА, ДЕЛИКТЕ ИЛИ ИНОЙ СИТУАЦИИ,
    ВОЗНИКШИМ ИЗ-ЗА ИСПОЛЬЗОВАНИЯ ПРОГРАММНОГО ОБЕСПЕЧЕНИЯ ИЛИ ИНЫХ
    ДЕЙСТВИЙ С ПРОГРАММНЫМ ОБЕСПЕЧЕНИЕМ.

 **********************************************************************}

{$I uev3defines.inc}

unit rubiroSensors;
interface
uses uev3, uev3Devices,uev3sysfs;
type
  TEv3SMode=(
     smUnknown,
     // цвето-свето-датчик
     smREFLECT,smAMBIENT,smCOLOR,smREFRAW,smRGBRAW,smCOLCAL,
     // nxt-свето-датчик
     smNREFLECT,smNAMBIENT,
     // ультра-датчик
     smUSDISTCM,smUSDISTIN,smUSLISTEN,smUSSICM,smUSSIIN,smUSDCCM,smUSDCIN,
     // гиро-датчик
     smGYROANG,smGYRORATE,smGYROFAS,smGYROGA,smGYROCAL,smTILTRATE,smTILTANG,
     // инфра-датчик
     smIRPROX,smIRSEEK,smIRREMOTE,smIRREMA,smIRSALT,smIRCAL
     );

  TEv3IRButton=(irNone,irRU,irRD,irBU,irBD,irRUBU,irRUBD,irRDBU,irRDBD,irBeacon,irRURD,irBUBD);
  TEv3IRButtons=set of TEv3IRButton;

const
  ev3SModes:array[TEv3SMode] of string[20]=(
      '',
      // цвето-свето-датчик
      'COL-REFLECT', 'COL-AMBIENT', 'COL-COLOR', 'REF-RAW', 'RGB-RAW', 'COL-CAL',
      // nxt-свето-датчик
      'REFLECT', 'AMBIENT',
      // ультра-датчик
      'US-DIST-CM', 'US-DIST-IN', 'US-LISTEN', 'US-SI-CM', 'US-SI-IN', 'US-DC-CM', 'US-DC-IN',
      // гиро-датчик
      'GYRO-ANG', 'GYRO-RATE', 'GYRO-FAS', 'GYRO-G&A', 'GYRO-CAL', 'TILT-RATE', 'TILT-ANG',
      // инфра-датчик
      'IR-PROX', 'IR-SEEK', 'IR-REMOTE', 'IR-REM-A', 'IR-S-ALT', 'IR-CAL'
      );


type

  { TEv3Sensor }
  TEv3SensorClass=class of TEv3Sensor;
  TEv3Sensor=class(TEv3OneDeviceHelper)
   private
    fmode:TEv3SMode;
    fwaitMode:boolean;
    function getMode: TEv3SMode;virtual;
    procedure setMode(AValue: TEv3SMode);virtual;
   protected
    // v0.2.5
    procedure onAfterSetMode(oldmode,newmode:TEv3SMode);virtual;
   public
    class function varToMode(s:variant):TEv3SMode;
    function value:integer;
    function value0:integer;
    function value1:integer;
    function value2:integer;
    function value3:integer;
    function value4:integer;
    function value5:integer;
    function value6:integer;
    function value7:integer;
    property mode:TEv3SMode read getMode write setMode;
    procedure setIntervals(R,W:integer);override;
    // Режим ожидания нового значения
    function waitMode(yes:variant):TEv3Sensor;
  end;

  { TEv3InfraSensor }

  TEv3InfraSensorClass=class of TEv3InfraSensor;
  TEv3InfraSensor=class(TEv3Sensor)
   private
     procedure setMode(AValue: TEv3SMode);override;
   protected
    class function deviceType:TEv3DeviceType;override;
     procedure onAfterSetMode(oldmode,newmode:TEv3SMode);override;
   public
    function waitMode(yes:variant):TEv3InfraSensor;
   {
    После переключения на новый режим, датчик требует реинициализации маяка.
    В результате практически невозможно программировать датчик
    со смешением режимов, например - ОДНОВРЕМЕННО получать и обрабатывать
    1) расстояние и направление до маяка,
    2) наборы кнопок на маяке
   }
    procedure setIntervals(R,W:integer);override;
    // расстояние до препятствия в %, до 0.7 метра
    function len:variant;
    function len(out ln:variant):TEv3InfraSensor;
    {
     направление до маяка, от -25 (слева) до +25 (справа)
     канал - от 1 до 4
    }
    function direct(const channel:variant):variant;
    function direct(const channel:variant; out dir:variant):TEv3InfraSensor;
    {
     расстояние до маяка в %, до 2 метров
     канал - от 1 до 4
    }
    function distance(const channel:variant):variant;
    function distance(const channel:variant; out dist:variant):TEv3InfraSensor;
    {
      нажатые на маяке кнопки, при использовании первого канала;
      любые комбинации кнопок irRU, irRD, irBU и irBD
    }
    function anyButtons():TEv3IRButtons;
    function anyButtons(out butt:TEv3IRButtons):TEv3InfraSensor;
    {
      нажатые на маяке кнопки, при использовании канала от 1 до 4;
      возвращает факт нажатия одной кнопки или комбинации двух кнопок.
      может анализироваться как значение типа TEv3IRButton,
      либо как целое значение из диапазона от 0 до 11
    }
    function buttons(const channel:variant):variant;
    function buttons(const channel:variant; out butt:variant):TEv3InfraSensor;
   end;

  { TEv3GyroSensor }
  TEv3GyroDrift=(drAxis,drIntegral);
  TEv3GyroSensorClass=class of TEv3GyroSensor;
  TEv3GyroSensor=class(TEv3Sensor)
   private
   // 0.2.6
   { 3 вида смещения (дрейфа)
     1. timeDrift - постоянно нарастающее смещение в состоянии покоя.
     Возникает, если подключение или калибровка гироскопа произведены
     НЕ В СОСТОЯНИИ ПОКОЯ. Однозначно детектируется ненулевой скоростью изменения
     угла поворота. Компенсация возможна, но сложна и при корректной калибровке
     не нужна.
     2. integralDrift - интегральное смещение, возникающее при вращении гироскопа.
     Величина пропорциональна количеству градусов, на которые поворачивался гироскоп
     с момента инициализации. Причины - погрешности в АЦП.
     Индивидуальная величина для каждого гироскопа. Компенсация возможна.
     3. axisDrift - осевое смещение, его величина определяется отклонением оси
     вращения от оси гироскопа, например - при наклонном закреплении
     датчика на роботе. Компенсация возможна, если смещение оси носит постоянный
     характер и не изменяется со временем. Постоянство смещения оси труднодостижимо
     в случае активных манипуляций с роботом
     (вибрации, наклоны при резких поворотах и т.п.).

     В текущей версии библиотеки возможна компенсация смещений
     integralDrift и axisDrift путем
     задания дрифт-коэффициентов, т.е. смещения на 1 градус.
   }
    fdriftRate:array[TEv3GyroDrift] of double; // все коэффициенты смещения
    fcorrection:array[TEv3GyroDrift] of double; // текущая коррекция смещения
    fBaseAngle:double; // начальный угол
    fprevAngle:integer; // предыдущее значение угла
   protected
    class function deviceType:TEv3DeviceType;override;
    procedure deviceRefreshed(dev:TEv3Device);override;
    procedure onAfterSetMode(oldmode,newmode:TEv3SMode);override;
   public
    procedure afterConstruction;override;
   public
    procedure setIntervals(R,W:integer);override;

    // 0.2.6 Установка-получение базовой угловой позиции
    function setBaseAngle(baseAngle:variant):TEv3GyroSensor;
    function baseAngle:variant;
    // 0.2.6 Увеличение значения базовой угловой позиции
    function addBaseAngle(addAngle:variant):TEv3GyroSensor;

    // 0.2.6 Установка-получение коэффициентов смещения
    function setDriftRate(drifttype:TEv3GyroDrift; driftValue:double):TEv3GyroSensor;
    function driftRate(drifttype:TEv3GyroDrift):double;
    property axisDriftRate:double read fdriftRate[drAxis] write fdriftRate[drAxis];
    property integralDriftRate:double read fdriftRate[drIntegral] write fdriftRate[drIntegral];

    // 0.2.6 Получение актуальной коррекции смещения указанного типа
    function Correction(drifttype:TEv3GyroDrift):double;
    property axisCorrection:double read fCorrection[drAxis];
    property integralCorrection:double read fCorrection[drIntegral];

    // калибровка гироскопа, должна проводится в состоянии покоя,
    // действует только на угол/скорость поворота,
    // не оказывает влияния на угол/скорость наклона
    function calibrate:TEv3GyroSensor;
    // угол поворота в градусах
    // 0.2.6
    // если src=true, то результат возвращается непосредственно с гироскопа
    // если src=false, то результат возвращается с учетом
    // базового угла и всех коррекций смещения
    function angle(src:boolean=false):variant;
    function angle(out ang:variant;src:boolean=false):TEv3GyroSensor;
    // скорость поворота
    function speed:variant;
    function speed(out sp:variant):TEv3GyroSensor;
    // угол наклона в градусах
    function tiltAngle:variant;
    function tiltAngle(out ang:variant):TEv3GyroSensor;
    // скорость наклона
    function tiltSpeed:variant;
    function tiltSpeed(out sp:variant):TEv3GyroSensor;
  end;


  { TEv3UltraSensor }
  TEv3UltraSensorClass=class of TEv3UltraSensor;
  TEv3UltraSensor=class(TEv3Sensor)
    private
     //prevData:cint;
     //prevTimeSingleMode:tDateTime;
     fsingleMode:boolean;
     function endIntervalSingleMode:boolean;
     procedure setMode(AValue: TEv3SMode);override;
     const fIntervalSingleMode=300;
    protected
     class function deviceType:TEv3DeviceType;override;
     procedure deviceRefreshed(dev:TEv3Device);override;
     procedure onAfterSetMode(oldmode,newmode:TEv3SMode);override;
    public
     {
      Включает (1/true) или отключает (0/false) режим ожидания
      нового расстояния. В этом режиме получение расстояния до
      препятствия блокирует программу до тех пор, пока оно не
      изменится по сравнения с предыдущим замером. По умолчанию
      отключено.
     }
     function waitMode(yes:variant):TEv3UltraSensor;
     // Режим единичного чтения, интервал получения - не менее 250-300ms (fIntervalSingleMode)
     // нужен при наличии нескольких датчиков во избежание помех
     // Внимание! В случае низкого напряжения, а также плохой погоды за окном
     // гарантируются непериодические сбои в модуле ядра при попытке установки данного режима!
     // Следите за звуковыми сигналами! Система до countMaxRefreshFactory раз будет
     // пытаться повторить установку режима, после чего честно сгенерирует Exception.
     function singleMode(yes:variant):TEv3UltraSensor;
     // определение помех, возвращает истину, если рядом другой мешающий сенсор
     function noise:variant;
     function noise(out yesnoise:variant):TEv3UltraSensor;
     // расстояние до препятствия в миллиметрах
     function lenMM:variant;
     function lenMM(out len:variant):TEv3UltraSensor;
     // расстояние до препятствия в сантиметрах
     function lenCM:variant;
     function lenCM(out len:variant):TEv3UltraSensor;
     // расстояние до препятствия в метрах
     function lenM:variant;
     function lenM(out len:variant):TEv3UltraSensor;

  end;


  { TEv3ColorSensor }
  TEv3ColorSensorClass=class of TEv3ColorSensor;
  TEv3ColorSensor=class(TEv3Sensor)
    public
     procedure setIntervals(R,W:integer);override;
    protected
     class function deviceType:TEv3DeviceType;override;
     procedure deviceRefreshed(dev:TEv3Device);override;
     procedure onAfterSetMode(oldmode,newmode:TEv3SMode);override;
    public
     {
     Включает (1/true) или отключает (0/false) режим ожидания
     нового значения цвета/света. В этом режиме получение значения цвета/света
     блокирует программу до тех пор, пока оно не
     изменится по сравнения с предыдущим замером. По умолчанию
     отключено.
     }
     function waitMode(yes:variant):TEv3ColorSensor;
     // получение цвета от 0 до 7
     function color:variant;
     function color(out col:variant):TEv3ColorSensor;

     // получение RGB цветов, каждый в диапазоне от 0 до 255
     function color(out colRed,colGreen,colBlue:variant):TEv3ColorSensor;
     function colorRed:variant;
     function colorRed(out colRed:variant):TEv3ColorSensor;
     function colorGreen:variant;
     function colorGreen(out colGreen:variant):TEv3ColorSensor;
     function colorBlue:variant;
     function colorBlue(out colBlue:variant):TEv3ColorSensor;

     // получение яркости отраженного света, от 0 до 100
     function reflect:variant;
     function reflect(out ref:variant):TEv3ColorSensor;
     // получение яркости освещения, от 0 до 100
     function ambient:variant;
     function ambient(out amb:variant):TEv3ColorSensor;
  end;



  { TNXTLightSensor }
  // 0.2.6
  TNXTLightSensorClass=class of TNXTLightSensor;
  TNXTLightSensor=class(TEv3Sensor)
    public
    protected
     class function deviceType:TEv3DeviceType;override;
     procedure deviceRefreshed(dev:TEv3Device);override;
    public
     // получение яркости отраженного света, от 0 до 100, вещественное значение
     function reflect:variant;
     function reflect(out ref:variant):TNXTLightSensor;
     // получение яркости освещения, от 0 до 100, вещественное значение
     function ambient:variant;
     function ambient(out amb:variant):TNXTLightSensor;
  end;


  { TEv3TouchSensor }
  TEv3TouchSensorClass=class of TEv3TouchSensor;
  TEv3TouchSensor=class(TEv3Sensor)
   protected
    class function deviceType:TEv3DeviceType;override;
   public
    // Истина при нажатой кнопке
    function pressed:boolean;
    function pressed(out yes:boolean):TEv3TouchSensor;
    // Ожидание нажатия кнопки
    function waitPress:TEv3TouchSensor;
    // Ожидание отпускания кнопки
    function waitRelease:TEv3TouchSensor;
    // Ожидание щелчка кнопки
    function waitClick:TEv3TouchSensor;
  end;

  //+++++++++++++++++++++++++++++++++++++++  02.04.18 
   TNXTTouchSensorClass=class of TNXTTouchSensor;
   TNXTTouchSensor=class(TEv3Sensor)
   protected
    class function deviceType:TEv3DeviceType;override;
   public
    function pressed:boolean;
    function pressed(out yes:boolean):TNxtTouchSensor;
    function waitPress:TNxtTouchSensor;
    function waitRelease:TNxtTouchSensor;
    function waitClick:TNxtTouchSensor;
  end;
  //+++++++++++++++++++++++++++++++++++++++

  {
   номерные функции возвращают датчик по порядку подключения вне зависимости
   от номера порта. Либо nil.

   функции с числовым параметром возвращают датчик по номеру порта (от 1 до 4). Либо nil.

   Функции с параметром типа TObject или TVarRec возвращают преобразованный к нужному типу
   объект, либо генерируют исключительную ситуацию, если объект несовместимого типа.
  }

    function nxtTouch1:TNXTTouchSensor;
    function nxtTouch2:TNXTTouchSensor;
    function nxtTouch3:TNXTTouchSensor;
    function nxtTouch4:TNXTTouchSensor;
    function nxtTouch(index:integer=-1):TNXTTouchSensor;
    function nxtTouch(obj:TVarRec):TNXTTouchSensor;
    function nxtTouch(obj:TObject):TNXTTouchSensor;

    function ev3Touch1:TEv3TouchSensor;
    function ev3Touch2:TEv3TouchSensor;
    function ev3Touch3:TEv3TouchSensor;
    function ev3Touch4:TEv3TouchSensor;
    function ev3Touch(index:integer=-1):TEv3TouchSensor;
    function ev3Touch(obj:TVarRec):TEv3TouchSensor;
    function ev3Touch(obj:TObject):TEv3TouchSensor;

    function ev3Color1:TEv3ColorSensor;
    function ev3Color2:TEv3ColorSensor;
    function ev3Color3:TEv3ColorSensor;
    function ev3Color4:TEv3ColorSensor;
    function ev3Color(index:integer=-1):TEv3ColorSensor;
    function ev3Color(obj:TVarRec):TEv3ColorSensor;
    function ev3Color(obj:TObject):TEv3ColorSensor;

    function nxtLight1:TNXTLightSensor;
    function nxtLight2:TNXTLightSensor;
    function nxtLight3:TNXTLightSensor;
    function nxtLight4:TNXTLightSensor;
    function nxtLight(index:integer=-1):TNXTLightSensor;
    function nxtLight(obj:TVarRec):TNXTLightSensor;
    function nxtLight(obj:TObject):TNXTLightSensor;


    function ev3infra1:TEv3infraSensor;
    function ev3infra2:TEv3infraSensor;
    function ev3infra3:TEv3infraSensor;
    function ev3infra4:TEv3infraSensor;
    function ev3infra(index:integer=-1):TEv3infraSensor;
    function ev3infra(obj:TVarRec):TEv3infraSensor;
    function ev3infra(obj:TObject):TEv3infraSensor;

    function ev3Gyro1:TEv3GyroSensor;
    function ev3Gyro2:TEv3GyroSensor;
    function ev3Gyro3:TEv3GyroSensor;
    function ev3Gyro4:TEv3GyroSensor;
    function ev3Gyro(index:integer=-1):TEv3GyroSensor;
    function ev3Gyro(obj:TVarRec):TEv3GyroSensor;
    function ev3Gyro(obj:TObject):TEv3GyroSensor;

    function ev3Ultra1:TEv3UltraSensor;
    function ev3Ultra2:TEv3UltraSensor;
    function ev3Ultra3:TEv3UltraSensor;
    function ev3Ultra4:TEv3UltraSensor;
    function ev3Ultra(index:integer=-1):TEv3UltraSensor;
    function ev3Ultra(obj:TVarRec):TEv3UltraSensor;
    function ev3Ultra(obj:TObject):TEv3UltraSensor;

//    function ev3Sensor(index: integer; sensorClass:TEv3SensorClass; refreshFactory:boolean=false): TEv3Sensor;

procedure outSet(s:TEv3IRButtons);

implementation
uses
  SysUtils,Unix,variants,dateutils;

__WARNGLOBALOFF

function nxtTouch1: TNXTTouchSensor;
begin
 result:=nxtTouch(-1);
end;

function nxtTouch2: TNXTTouchSensor;
begin
 result:=nxtTouch(-2);
end;

function nxtTouch3: TNXTTouchSensor;
begin
 result:=nxtTouch(-3);
end;

function nxtTouch4: TNXTTouchSensor;
begin
 result:=nxtTouch(-4);
end;

function nxtTouch(index: integer): TNXTTouchSensor;
begin
 result:=ev3Helper(index,TNXTTouchSensor) as TNXTTouchSensor;
end;

function nxtTouch(obj: TVarRec): TNXTTouchSensor;
begin
 result:=nxtTouch(obj.VObject);
end;

function nxtTouch(obj: TObject): TNXTTouchSensor;
begin
 result:=obj as TNXTTouchSensor;
end;

function ev3Touch1: TEv3TouchSensor;
begin
 result:=ev3Touch(-1);
end;

function ev3Touch2: TEv3TouchSensor;
begin
 result:=ev3Touch(-2);
end;

function ev3Touch3: TEv3TouchSensor;
begin
 result:=ev3Touch(-3);
end;

function ev3Touch4: TEv3TouchSensor;
begin
 result:=ev3Touch(-4);
end;

function ev3Touch(index: integer): TEv3TouchSensor;
begin
 result:=ev3Helper(index,TEv3TouchSensor) as TEv3TouchSensor;
end;

function ev3Touch(obj: TVarRec): TEv3TouchSensor;
begin
 result:=TEv3TouchSensor(obj.VObject);
end;

function ev3Touch(obj: TObject): TEv3TouchSensor;
begin
 result:=obj as TEv3TouchSensor;
end;

//--------------------
function ev3Color1: TEv3ColorSensor;
begin
 result:=ev3Color(-1);
end;

function ev3Color2: TEv3ColorSensor;
begin
 result:=ev3Color(-2);
end;

function ev3Color3: TEv3ColorSensor;
begin
 result:=ev3Color(-3);
end;

function ev3Color4: TEv3ColorSensor;
begin
 result:=ev3Color(-4);
end;

function ev3Color(index: integer): TEv3ColorSensor;
begin
 result:=ev3Helper(index,TEv3ColorSensor) as TEv3ColorSensor;
end;

function ev3Color(obj: TVarRec): TEv3ColorSensor;
begin
 result:=ev3Color(obj.VObject);
end;

function ev3Color(obj: TObject): TEv3ColorSensor;
begin
 result:=obj as TEv3ColorSensor;
end;

//--------------------

function nxtLight1: TNXTLightSensor;
begin
 result:=nxtLight(-1);
end;

function nxtLight2: TNXTLightSensor;
begin
 result:=nxtLight(-2);
end;

function nxtLight3: TNXTLightSensor;
begin
 result:=nxtLight(-3);
end;

function nxtLight4: TNXTLightSensor;
begin
 result:=nxtLight(-4);
end;

function nxtLight(index: integer): TNXTLightSensor;
begin
  result:=ev3Helper(index,TNXTLightSensor) as TNXTLightSensor;
end;

function nxtLight(obj: TVarRec): TNXTLightSensor;
begin
 result:=nxtLight(obj.VObject);
end;

function nxtLight(obj: TObject): TNXTLightSensor;
begin
 result:=obj as TNXTLightSensor;
end;

//--------------------
function ev3Infra1: TEv3InfraSensor;
begin
 result:=ev3Infra(-1);
end;

function ev3Infra2: TEv3InfraSensor;
begin
 result:=ev3Infra(-2);
end;

function ev3Infra3: TEv3InfraSensor;
begin
 result:=ev3Infra(-3);
end;

function ev3Infra4: TEv3InfraSensor;
begin
 result:=ev3Infra(-4);
end;

function ev3Infra(index: integer): TEv3InfraSensor;
begin
 result:=ev3Helper(index,TEv3InfraSensor) as TEv3InfraSensor;
end;

function ev3infra(obj: TVarRec): TEv3infraSensor;
begin
 result:=ev3Infra(obj.VObject);
end;

function ev3infra(obj: TObject): TEv3infraSensor;
begin
 result:=obj as TEv3infraSensor;
end;

//--------------------
function ev3Ultra1: TEv3UltraSensor;
begin
 result:=ev3Ultra(-1);
end;

function ev3Ultra2: TEv3UltraSensor;
begin
 result:=ev3Ultra(-2);
end;

function ev3Ultra3: TEv3UltraSensor;
begin
 result:=ev3Ultra(-3);
end;

function ev3Ultra4: TEv3UltraSensor;
begin
 result:=ev3Ultra(-4);
end;

function ev3Ultra(index: integer): TEv3UltraSensor;
begin
 result:=ev3Helper(index,TEv3UltraSensor) as TEv3UltraSensor;
end;

function ev3Ultra(obj: TVarRec): TEv3UltraSensor;
begin
 result:=ev3ultra(obj.VObject);
end;

function ev3Ultra(obj: TObject): TEv3UltraSensor;
begin
 result:=obj as TEv3UltraSensor;
end;


//--------------------
function ev3Gyro1: TEv3GyroSensor;
begin
 result:=ev3Gyro(-1);
end;

function ev3Gyro2: TEv3GyroSensor;
begin
 result:=ev3Gyro(-2);
end;

function ev3Gyro3: TEv3GyroSensor;
begin
 result:=ev3Gyro(-3);
end;

function ev3Gyro4: TEv3GyroSensor;
begin
 result:=ev3Gyro(-4);
end;

function ev3Gyro(index: integer): TEv3GyroSensor;
begin
 result:=ev3Helper(index,TEv3GyroSensor) as TEv3GyroSensor;
end;

function ev3Gyro(obj: TVarRec): TEv3GyroSensor;
begin
 result:=ev3Gyro(obj.VObject);
end;

function ev3Gyro(obj: TObject): TEv3GyroSensor;
begin
 result:=obj as TEv3GyroSensor;
end;




procedure outSet(s:TEv3IRButtons);
 var i:TEv3IRButton;
begin
 for i in s do write(i,' ');
 writeln;
end;

{ TNXTLightSensor }

class function TNXTLightSensor.deviceType: TEv3DeviceType;
begin
 Result:=nxttLightSensor;
end;

procedure TNXTLightSensor.deviceRefreshed(dev: TEv3Device);
begin
  inherited deviceRefreshed(dev);
  dev.Attrs[ev3amode].sData:=ev3SModes[fmode];
end;

{
function TEv3ColorSensor.reflect: variant;
begin
 mode:=smREFLECT; result:=value0;
end;

function TEv3ColorSensor.reflect(out ref: variant): TEv3ColorSensor;
begin
 result:=self;
 ref:=self.reflect;
end;

function TEv3ColorSensor.ambient: variant;
begin
 mode:=smAMBIENT; result:=value0;
end;

function TEv3ColorSensor.ambient(out amb: variant): TEv3ColorSensor;
begin
 result:=self;
 amb:=self.ambient;
end;
}



function TNXTLightSensor.reflect: variant;
begin
 mode:=smNREFLECT; result:=value0/10;
end;

function TNXTLightSensor.reflect(out ref: variant): TNXTLightSensor;
begin
 result:=self;
 ref:=self.reflect;
end;

function TNXTLightSensor.ambient: variant;
begin
 mode:=smNAMBIENT; result:=value0/10;
end;

function TNXTLightSensor.ambient(out amb: variant): TNXTLightSensor;
begin
 result:=self;
 amb:=self.ambient;
end;

{ TEv3InfraSensor }

procedure TEv3InfraSensor.setIntervals(R, W: integer);
begin
  inherited setIntervals(R, W);
  ev.Attrs[ev3avalue1].setIntervals(R,W);
  ev.Attrs[ev3avalue2].setIntervals(R,W);
  ev.Attrs[ev3avalue3].setIntervals(R,W);
  ev.Attrs[ev3avalue4].setIntervals(R,W);
  ev.Attrs[ev3avalue5].setIntervals(R,W);
  ev.Attrs[ev3avalue6].setIntervals(R,W);
  ev.Attrs[ev3avalue7].setIntervals(R,W);
end;

function TEv3InfraSensor.len: variant;
begin
 mode:=smIRPROX;
 result:=value;
end;

function TEv3InfraSensor.len(out ln: variant): TEv3InfraSensor;
begin
 result:=self;
 ln:=self.len;
end;

function TEv3InfraSensor.direct(const channel: variant): variant;
begin
 mode:=smIRSEEK;
 case varToStr(channel) of
  '1': result:=value0;
  '2': result:=value2;
  '3': result:=value4;
  '4': result:=value6;
  else result:=value0;
 end;
end;

function TEv3InfraSensor.direct(const channel: variant; out dir: variant
  ): TEv3InfraSensor;
begin
 result:=self;
 dir:=self.direct(channel);
end;

function TEv3InfraSensor.distance(const channel: variant): variant;
begin
 mode:=smIRSEEK;
 case varToStr(channel) of
  '1': result:=value1;
  '2': result:=value3;
  '3': result:=value5;
  '4': result:=value7;
  else result:=value1;
 end;
end;

function TEv3InfraSensor.distance(const channel: variant; out dist: variant
  ): TEv3InfraSensor;
begin
 result:=self;
 dist:=self.distance(channel);
end;

function TEv3InfraSensor.anyButtons: TEv3IRButtons;
 var v:word;s:TEv3IRButtons;
begin
 mode:=smIRREMA;
 v:=value0;
 s:=[];
 if (v and $f)=0 then exit(s);
 if (v and $80)<>0 then include(s,irBD);
 if (v and $40)<>0 then include(s,irBU);
 if (v and $20)<>0 then include(s,irRD);
 if (v and $10)<>0 then include(s,irRU);
 result:=s;
end;

function TEv3InfraSensor.anyButtons(out butt: TEv3IRButtons): TEv3InfraSensor;
begin
 result:=self;
 butt:=self.anyButtons;
end;

function TEv3InfraSensor.buttons(const channel: variant): variant;
begin
 mode:=smIRREMOTE;
 case varToStr(channel) of
  '1': result:=value0;
  '2': result:=value1;
  '3': result:=value2;
  '4': result:=value3;
  else result:=value0;
 end;
end;

function TEv3InfraSensor.buttons(const channel: variant; out butt: variant
  ): TEv3InfraSensor;
begin
 result:=self;
 butt:=self.buttons(channel);
end;

procedure TEv3InfraSensor.setMode(AValue: TEv3SMode);
begin
  inherited setMode(AValue);
{
 if aValue=fmode then exit;
 fmode:=avalue;
// writeln('+mode: ',fmode);
 ev.Attrs[ev3amode].sData:=ev3SModes[fmode];
 sleep(1000);
 //writeln('-mode: ',fmode);
}
end;

class function TEv3InfraSensor.deviceType: TEv3DeviceType;
begin
 result:=ev3tInfraSensor;
end;

procedure TEv3InfraSensor.onAfterSetMode(oldmode, newmode: TEv3SMode);
begin
  inherited onAfterSetMode(oldmode, newmode);
  ev.Attrs[ev3avalue1].reInit;
  ev.Attrs[ev3avalue2].reInit;
  ev.Attrs[ev3avalue3].reInit;
  ev.Attrs[ev3avalue4].reInit;
  ev.Attrs[ev3avalue5].reInit;
  ev.Attrs[ev3avalue6].reInit;
  ev.Attrs[ev3avalue7].reInit;
end;

function TEv3InfraSensor.waitMode(yes: variant): TEv3InfraSensor;
begin
  result:=self;
  inherited waitMode(yes);
end;



{ TEv3GyroSensor }

class function TEv3GyroSensor.deviceType: TEv3DeviceType;
begin
  Result:=ev3tGyroSensor;
end;

procedure TEv3GyroSensor.deviceRefreshed(dev: TEv3Device);
begin
  inherited deviceRefreshed(dev);
  dev.Attrs[ev3amode].sData:=ev3SModes[fmode];
end;

procedure TEv3GyroSensor.afterConstruction;
begin
  mode:=smGYROGA;
  fprevAngle:=value0;
  fBaseAngle:=0;
  fillbyte(fCorrection,sizeof(fcorrection),0);
  fillbyte(fdriftRate,sizeof(fdriftRate),0);
end;

procedure TEv3GyroSensor.onAfterSetMode(oldmode, newmode: TEv3SMode);
begin
  inherited onAfterSetMode(oldmode, newmode);
  ev.Attrs[ev3avalue1].reInit();
  fprevAngle:=0;
  fillbyte(fCorrection,sizeof(fcorrection),0);
end;


procedure TEv3GyroSensor.setIntervals(R, W: integer);
begin
 inherited setIntervals(R,W);
 ev.Attrs[ev3avalue1].setIntervals(R,W);
end;

function TEv3GyroSensor.setBaseAngle(baseAngle: variant): TEv3GyroSensor;
begin
  fBaseAngle:=baseAngle;
  result:=self;
end;

function TEv3GyroSensor.baseAngle: variant;
begin
 result:=fBaseAngle;
end;

function TEv3GyroSensor.addBaseAngle(addAngle: variant): TEv3GyroSensor;
begin
  result:=self;
  fBaseAngle+=addAngle;
end;

function TEv3GyroSensor.setDriftRate(drifttype: TEv3GyroDrift; driftValue: double
  ): TEv3GyroSensor;
begin
 result:=self;
 fdriftRate[driftType]:=driftvalue;
end;

function TEv3GyroSensor.driftRate(drifttype: TEv3GyroDrift): double;
begin
 result:=fdriftRate[driftType];
end;

function TEv3GyroSensor.Correction(drifttype: TEv3GyroDrift): double;
begin
 result:=fCorrection[driftType];
end;

function TEv3GyroSensor.calibrate: TEv3GyroSensor;
 var _mode:TEv3SMode;// i:TEv3GyroDrift;
begin
 // writeln('---- ',fmode);
  result:=self;
  _mode:=fmode;
  //if (fmode=smGYRORATE) then mode:=smGYROGA else mode:=smGYRORATE;
  mode:=smGYROCAL;
  mode:=_mode;
  fprevAngle:=0;
  fillbyte(fCorrection,sizeof(fcorrection),0);
end;

function TEv3GyroSensor.angle(src:boolean=false): variant;
begin
 if (fmode<>smGYROANG)and(fmode<>smGYROGA) then begin
  mode:=smGYROGA;
 end;
 result:=value0;
 fCorrection[drAxis]:=result*fdriftRate[drAxis];
 fCorrection[drIntegral]+=abs(result-fPrevAngle)*fdriftRate[drIntegral];
 fprevAngle:=result;
 if src then exit;
 // интегральная часть всегда вычитается
 // осевая часть вычитается из положительной величины и добавляется к отрицательной
 if result>=0 then
  result:=result-fBaseAngle-fCorrection[drAxis]-fCorrection[drIntegral]
 else
  result:=result-fBaseAngle+fCorrection[drAxis]-fCorrection[drIntegral];
end;

function TEv3GyroSensor.angle(out ang: variant; src: boolean): TEv3GyroSensor;
begin
  result:=self;
  ang:=self.angle(src);
end;

function TEv3GyroSensor.speed: variant;
begin
 if fmode=smGYRORATE then exit(value0);
 if fmode=smGYROGA then exit(value1);
 mode:=smGYROGA;
 exit(speed);
end;

function TEv3GyroSensor.speed(out sp: variant): TEv3GyroSensor;
begin
 result:=self;
 sp:=self.speed;
end;

function TEv3GyroSensor.tiltAngle: variant;
begin
 if (fmode<>smTILTANG) then mode:=smTILTANG;
 result:=value0;
end;

function TEv3GyroSensor.tiltAngle(out ang: variant): TEv3GyroSensor;
begin
  result:=self;
  ang:=self.tiltAngle;
end;

function TEv3GyroSensor.tiltSpeed: variant;
begin
 if (fmode<>smTILTRATE) then mode:=smTILTRATE;
 result:=value0;
end;

function TEv3GyroSensor.tiltSpeed(out sp: variant): TEv3GyroSensor;
begin
 result:=self;
 sp:=self.tiltSpeed;

end;

{ TEv3Sensor }

function TEv3Sensor.getMode: TEv3SMode;
var dt:string;
begin
 dt:=ev.Attrs[ev3amode].sData;
 for fmode in Tev3SMode do
  if ev3SModes[fmode]=dt then exit(fmode);
end;

procedure TEv3Sensor.setMode(AValue: TEv3SMode);
 var oldmode:TEv3SMode;
begin
 if aValue=fmode then exit;
 oldmode:=fmode;
 fmode:=avalue;
// writeln('+mode: ',fmode);
 ev.Attrs[ev3amode].sData:=ev3SModes[fmode];
 onAfterSetMode(oldmode,avalue);
 //writeln('-mode: ',fmode);
end;


class function TEv3Sensor.varToMode(s: variant): TEv3SMode;
 var i:TEv3SMode;
begin
 for i:=low(i) to high(i) do begin
   if (i=s)or(ev3SModes[i]=s) then exit(i);
 end;
 result:=low(i);
end;

function TEv3Sensor.value: integer;
begin
 result:=value0;
end;

function TEv3Sensor.value0: integer;
begin
 if fWaitMode then result:=data.wiData
 else result:=data.iData;
end;

function TEv3Sensor.value1: integer;
begin
 if fWaitMode then result:=ev.Attrs[ev3avalue1].wiData
 else result:=ev.Attrs[ev3avalue1].iData;
end;

function TEv3Sensor.value2: integer;
begin
 if fWaitMode then result:=ev.Attrs[ev3avalue2].wiData
 else result:=ev.Attrs[ev3avalue2].iData;
end;

function TEv3Sensor.value3: integer;
begin
 if fWaitMode then result:=ev.Attrs[ev3avalue3].wiData
 else result:=ev.Attrs[ev3avalue3].iData;
end;

function TEv3Sensor.value4: integer;
begin
 if fWaitMode then result:=ev.Attrs[ev3avalue4].wiData
 else result:=ev.Attrs[ev3avalue4].iData;
end;

function TEv3Sensor.value5: integer;
begin
 if fWaitMode then result:=ev.Attrs[ev3avalue5].wiData
 else result:=ev.Attrs[ev3avalue5].iData;
end;

function TEv3Sensor.value6: integer;
begin
 if fWaitMode then result:=ev.Attrs[ev3avalue6].wiData
 else result:=ev.Attrs[ev3avalue6].iData;
end;

function TEv3Sensor.value7: integer;
begin
 if fWaitMode then result:=ev.Attrs[ev3avalue7].wiData
 else result:=ev.Attrs[ev3avalue7].iData;
end;

procedure TEv3Sensor.setIntervals(R, W: integer);
begin
  data.setIntervals(R,W);
  // --------
end;

function TEv3Sensor.waitMode(yes: variant): TEv3Sensor;
begin
 result:=self;
 fwaitMode:=varToYes(yes);
end;


__WARNOFFUNUSED
procedure TEv3Sensor.onAfterSetMode(oldmode, newmode: TEv3SMode);
begin
 data.reInit();
end;
__WARNON


{ TEv3UltraSensor }

function TEv3UltraSensor.endIntervalSingleMode: boolean;
//  var dt:TDateTime;
 begin
//  dt:=now();
//  result:=MilliSecondsBetween(prevTimeSingleMode,dt)>fIntervalSingleMode;
//  if result then prevTimeSingleMode:=dt;
// v0.2
  result:=_timer(9)>fIntervalSingleMode;
  if result then _timerStart(9);
 end;

procedure TEv3UltraSensor.setMode(AValue: TEv3SMode);
 var oldmode:TEv3SMode;
begin
  if (aValue=smUSSICM) and not endIntervalSingleMode then exit;
  if (aValue=fmode) and (aValue<>smUSSICM) then exit;
//  if avalue=smUSSICM then writeln('++++') else writeln('----+-----');
  oldmode:=fmode;
  fmode:=avalue;
  ev.Attrs[ev3amode].sData:=ev3SModes[fmode];
  onAfterSetMode(oldmode,avalue);
end;

class function TEv3UltraSensor.deviceType: TEv3DeviceType;
begin
  Result:=ev3tUltraSensor;
end;


procedure TEv3UltraSensor.deviceRefreshed(dev: TEv3Device);
begin
  inherited deviceRefreshed(dev);
  dev.Attrs[ev3amode].sData:=ev3SModes[fmode];
end;

procedure TEv3UltraSensor.onAfterSetMode(oldmode, newmode: TEv3SMode);
begin
  inherited onAfterSetMode(oldmode, newmode);
end;


function TEv3UltraSensor.waitMode(yes: variant): TEv3UltraSensor;
begin
 result:=self;
 inherited waitmode(yes);
end;

function TEv3UltraSensor.singleMode(yes: variant): TEv3UltraSensor;
begin
 result:=self;
 fsingleMode:=varToYes(yes);
end;

function TEv3UltraSensor.lenMM: variant;
begin
 if fSingleMode then mode:=smUSSICM
 else mode:=smUSDISTCM;
 result:=value;
end;

function TEv3UltraSensor.lenMM(out len: variant): TEv3UltraSensor;
begin
 result:=self;
 len:=self.lenMM;
end;

function TEv3UltraSensor.lenCM: variant;
begin
 result:=lenmm/10;
end;

function TEv3UltraSensor.lenCM(out len: variant): TEv3UltraSensor;
begin
 result:=self;
 len:=lenmm/10;
end;

function TEv3UltraSensor.lenM: variant;
begin
 result:=lenmm/100;
end;

function TEv3UltraSensor.lenM(out len: variant): TEv3UltraSensor;
begin
 result:=self;
 len:=lenmm/100;
end;

function TEv3UltraSensor.noise: variant;
begin
  mode:=smUSLISTEN;
  result:=boolean(value);
end;

function TEv3UltraSensor.noise(out yesnoise: variant): TEv3UltraSensor;
begin
 result:=self;
 yesnoise:=self.noise;
end;

{ TEv3ColorSensor }
function TEv3ColorSensor.waitMode(yes: variant): TEv3ColorSensor;
begin
 result:=self;
 inherited waitmode(yes);
end;

function TEv3ColorSensor.color: variant;
begin
  mode:=smCOLOR;result:=value0;
end;

function TEv3ColorSensor.color(out col: variant): TEv3ColorSensor;
begin
  result:=self;
  col:=self.color;
end;

function TEv3ColorSensor.color(out colRed, colGreen, colBlue: variant
  ): TEv3ColorSensor;
begin
 result:=self;
 colRed:=self.colorRed;
 colGreen:=self.colorGreen;
 colBlue:=self.colorBlue;
end;

function TEv3ColorSensor.colorRed: variant;
begin
  mode:=smRGBRAW; result:=value0;
end;

function TEv3ColorSensor.colorRed(out colRed: variant): TEv3ColorSensor;
begin
 result:=self;
 colRed:=self.colorRed;
end;

function TEv3ColorSensor.colorGreen: variant;
begin
 mode:=smRGBRAW; result:=value1;
end;

function TEv3ColorSensor.colorGreen(out colGreen: variant): TEv3ColorSensor;
begin
 result:=self;
 colGreen:=self.colorGreen;
end;

function TEv3ColorSensor.colorBlue: variant;
begin
 mode:=smRGBRAW; result:=value2;
end;

function TEv3ColorSensor.colorBlue(out colBlue: variant): TEv3ColorSensor;
begin
 result:=self;
 colBlue:=self.colorBlue;
end;

function TEv3ColorSensor.reflect: variant;
begin
 mode:=smREFLECT; result:=value0;
end;

function TEv3ColorSensor.reflect(out ref: variant): TEv3ColorSensor;
begin
 result:=self;
 ref:=self.reflect;
end;

function TEv3ColorSensor.ambient: variant;
begin
 mode:=smAMBIENT; result:=value0;
end;

function TEv3ColorSensor.ambient(out amb: variant): TEv3ColorSensor;
begin
 result:=self;
 amb:=self.ambient;
end;

procedure TEv3ColorSensor.onAfterSetMode(oldmode, newmode: TEv3SMode);
begin
  inherited onAfterSetMode(oldmode, newmode);
  if newMode=smRGBRAW then begin
    ev.Attrs[ev3avalue1].reInit();
    ev.Attrs[ev3avalue2].reInit();
  end;
end;

procedure TEv3ColorSensor.setIntervals(R, W: integer);
begin
 inherited setIntervals(R,W);
 ev.Attrs[ev3avalue1].setIntervals(R,W);
 ev.Attrs[ev3avalue2].setIntervals(R,W);
// ev.Attrs[ev3avalue3].setIntervals(R,W);
end;

class function TEv3ColorSensor.deviceType: TEv3DeviceType;
begin
  Result:=ev3tColorSensor;
end;

procedure TEv3ColorSensor.deviceRefreshed(dev: TEv3Device);
begin
  inherited deviceRefreshed(dev);
  dev.Attrs[ev3amode].sData:=ev3SModes[fmode];
end;

{ TEv3TouchSensor }

class function TEv3TouchSensor.deviceType: TEv3DeviceType;
begin
  Result:=ev3tTouchSensor;
end;

function TEv3TouchSensor.pressed: boolean;
begin
  result:=data.iData=1;
end;

function TEv3TouchSensor.pressed(out yes: boolean): TEv3TouchSensor;
begin
  result:=self;
  yes:=self.pressed;
end;

function TEv3TouchSensor.waitPress: TEv3TouchSensor;
begin
 result:=self;
 while data.wiData<>1 do;
end;

function TEv3TouchSensor.waitRelease: TEv3TouchSensor;
begin
 result:=self;
 while data.wiData<>0 do;
end;

function TEv3TouchSensor.waitClick: TEv3TouchSensor;
begin
 result:=self;
 waitPress.waitRelease;
end;

{ TNxtTouchSensor }

class function TNxtTouchSensor.deviceType: TEv3DeviceType;
begin
  Result:=nxttTouchSensor;
end;

function TNxtTouchSensor.pressed: boolean;
begin
  result:=data.iData=1;
end;

function TNxtTouchSensor.pressed(out yes: boolean): TNxtTouchSensor;
begin
  result:=self;
  yes:=self.pressed;
end;

function TNxtTouchSensor.waitPress: TNxtTouchSensor;
begin
 result:=self;
 while data.wiData<>1 do;
end;

function TNxtTouchSensor.waitRelease: TNxtTouchSensor;
begin
 result:=self;
 while data.wiData<>0 do;
end;

function TNxtTouchSensor.waitClick: TNxtTouchSensor;
begin
 result:=self;
 waitPress.waitRelease;
end;


//----------------------------------------------
procedure Done;
  var ev:TEv3Device;
 begin
  //WD('Sensor Done 1');
  if not assigned(ev3DeviceFactory) then exit;
  //WD('Sensor Done 2');
  for ev in ev3DeviceFactory do begin
    case ev.Device of
     ev3tColorSensor: begin
     // WD(ev.Attrs[ev3amode].sData);
      ev.Attrs[ev3amode].sData:=ev3SModes[smREFLECT];
     // WD(ev3SModes[smREFLECT]);
     // WD(ev.Attrs[ev3amode].sData);
     // WD('ColorSensor Reflect Set');
     // ev3.sleep(2000);
     end;
     ev3tUltraSensor: ev.Attrs[ev3amode].sData:=ev3SModes[smUSDISTCM];
     ev3tGyroSensor: ev.Attrs[ev3amode].sData:=ev3SModes[smGYROANG];
    end;
  end;
 end;

initialization
 pushDone(@Done);
end.

